import numpy as np
import matplotlib.pyplot as plt
from grid_world import standard_grid, negative_grid
import time

GAMMA = 0.9
ALPHA = 0.1
ALL_POSSIBLE_ACTIONS = ('R', 'U', 'D', 'L')
#np.random.seed(3547835)
#np.random.seed(72343232227)
fig = plt.figure(figsize=(11,9),facecolor='white')


def print_values(V, g):
  time.sleep(1)
  for j in reversed(range(g.height)):

    print ("'---------------------------------------")
   
    for i in range(g.width):
      v = V.get((i,j), 0)
      if v >= 0:
          print( "| ",'{:06.2f}'.format(v), end=" ")  

      else:
         print( "| ",'{:06.2f}'.format(v), end=" ") # -ve sign takes up an extra space

    print( "| ")
  print ("'---------------------------------------")
  print ("")
  print ("")



def print_policy(P, g):
  time.sleep(2)
  for j in reversed(range(g.height)):
    print ("-----------------------------")
    for i in range(g.width):
      a = P.get((i,j), ' ')
      # print ("  %s  |" % a)
      print( "| ",'{:3s}'.format(a),end=" ") 
    print( "| ")
  print ("-----------------------------")
  print ("")
  print ("")

def max_dict(d):
  # returns the argmax (key) and max (value) from a dictionary
  # put this into a function since we are using it so often
  max_key = None
  max_val = float('-inf')
  for k, v in d.items():
    if v > max_val:
      max_val = v
      max_key = k
  return max_key, max_val



def random_action(a, eps=0.1):
  # choose given a with probability 1 - eps + eps/4
  # choose some other a' != a with probability eps/4
  p = np.random.random()
  # if p < (1 - eps + eps/len(ALL_POSSIBLE_ACTIONS)):
  #   return a
  # else:
  #   tmp = list(ALL_POSSIBLE_ACTIONS)
  #   tmp.remove(a)
  #   return np.random.choice(tmp)
  #
  # this is equivalent to the above
  if p < (1 - eps):
    return a
  else:
    return np.random.choice(ALL_POSSIBLE_ACTIONS)




if __name__ == '__main__':

    grid = standard_grid()
    #grid = negative_grid(step_cost=-20)

    # print rewards
    #print "rewards:"
    print_values(grid.rewards, grid)

    # no policy initialization, we will derive our policy from most recent Q

    # initialize Q(s,a)
    Q = {}
    states = grid.all_states()
    for s in states:
        Q[s] = {}
        for a in ALL_POSSIBLE_ACTIONS:
            Q[s][a] = 0

    """
    # let's also keep track of how many times Q[s] has been updated
    update_counts = {}
    update_counts_sa = {}
    for s in states:
        update_counts_sa[s] = {}
        for a in ALL_POSSIBLE_ACTIONS:
            update_counts_sa[s][a] = 1.0
    """

    # repeat until convergence
    t = 1.0
    deltas = []
    for it in range(5000):
        if it % 100 == 0:
            t += 1e-2
        #if it % 2000 == 0:
            #print "it:", it

        # instead of 'generating' an epsiode, we will PLAY
        # an episode within this loop
        s = (3, 0) # start state
        grid.set_state(s)

        # the first (s, r) tuple is the state we start in and 0
        # (since we don't get a reward) for simply starting the game
        # the last (s, r) tuple is the terminal state and the final reward
        # the value for the terminal state is by definition 0, so we don't
        # care about updating it.
        a= np.random.choice(ALL_POSSIBLE_ACTIONS)
        #a = random_action(a, eps=0.1)
        biggest_change = 0
        #print 'old a is', a
        while not grid.game_over():

            a = random_action(a, eps=0.05) # epsilon-greedy
            # random action also works, but slower since you can bump into walls
            #a = np.random.choice(ALL_POSSIBLE_ACTIONS)

            r = grid.move(a)
            s2 = grid.current_state()
            a2, max_q_s2a2 = max_dict(Q[s2])

            #print 'a2 is', a2

            # we will update Q(s,a) AS we experience the episode
            #alpha = ALPHA / update_counts_sa[s][a]
            alpha = 0.1
            #update_counts_sa[s][a] += 0.005
            old_qsa = Q[s][a]
            Q[s][a] = Q[s][a] + alpha*(r + GAMMA*max_q_s2a2 - Q[s][a])

            biggest_change = max(biggest_change, np.abs(old_qsa - Q[s][a]))
            s = s2


            #print ' every time updated a is ',s, Q[s], Q[s][a]

        deltas.append(biggest_change)

    plt.plot(deltas)
    plt.show()

    # determine the policy from Q*
    # find V* from Q*
    policy = {}
    V = {}
    for s in grid.actions.keys():
        a, max_q = max_dict(Q[s])
        policy[s] = a
        V[s] = max_q
        #print_policy(policy, grid)
    #print ("policy and v", policy[s], V[s])
    # what's the proportion of time we spend updating each part of Q?
    """
    print "update counts:"
    total = np.sum(update_counts.values())
    for k, v in update_counts.iteritems():
        update_counts[k] = float(v) / total
    print_values(update_counts, grid)


    print "values:"
    print_values(V, grid)
    print "policy:"
    print_policy(policy, grid)
    """
