import numpy as np
import matplotlib.pyplot as plt
from grid_world import standard_grid, negative_grid
import time

GAMMA = 0.9
ALPHA = 0.1
ALL_POSSIBLE_ACTIONS = ('R', 'U', 'D', 'L')
#np.random.seed(3547835)
#np.random.seed(72343232227)


def print_values(V, g):
  time.sleep(1)
  for j in reversed(range(g.height)):

    print ("'---------------------------------------")
   
    for i in range(g.width):
      v = V.get((i,j), 0)
      if v >= 0:
          print( "| ",'{:06.2f}'.format(v), end=" ")  

      else:
         print( "| ",'{:06.2f}'.format(v), end=" ") # -ve sign takes up an extra space

    print( "| ")
  print ("'---------------------------------------")
  print ("")
  print ("")



def print_policy(P, g):
  time.sleep(2)
  for j in reversed(range(g.height)):
    print ("-----------------------------")
    for i in range(g.width):
      a = P.get((i,j), ' ')
      # print ("  %s  |" % a)
      print( "| ",'{:3s}'.format(a),end=" ") 
    print( "| ")
  print ("-----------------------------")
  print ("")
  print ("")

def max_dict(d):
  # returns the argmax (key) and max (value) from a dictionary
  # put this into a function since we are using it so often
  max_key = None
  max_val = float('-inf')
  for k, v in d.items():
    if v > max_val:
      max_val = v
      max_key = k
  return max_key, max_val

def random_action(a, eps=0.1):
  # choose given a with probability 1 - eps + eps/4
  # choose some other a' != a with probability eps/4
  p = np.random.random()
  # if p < (1 - eps + eps/len(ALL_POSSIBLE_ACTIONS)):
  #   return a
  # else:
  #   tmp = list(ALL_POSSIBLE_ACTIONS)
  #   tmp.remove(a)
  #   return np.random.choice(tmp)
  #
  # this is equivalent to the above
  if p < (1 - eps):
    return a
  else:
    return np.random.choice(ALL_POSSIBLE_ACTIONS)

if __name__ == '__main__':
    # NOTE: if we use the standard grid, there's a good chance we will end up with
    # suboptimal policies
    # e.g.
    # ---------------------------
    #   R  |   R  |   R  |      |
    # ---------------------------
    #   R* |      |   U  |      |
    # ---------------------------
    #   U  |   R  |   U  |   L  |
    # since going R at (1,0) (shown with a *) incurs no cost, it's OK to keep doing that.
    # we'll either end up staying in the same spot, or back to the start (2,0), at which
    # point we whould then just go back up, or at (0,0), at which point we can continue
    # on right.
    # instead, let's penalize each movement so the agent will find a shorter route.
    #
    grid = standard_grid()
    #grid = negative_grid(step_cost=-20)

    # print rewards
    #print "rewards:"
    print_values(grid.rewards, grid)

    # no policy initialization, we will derive our policy from most recent Q

    # initialize Q(s,a)
    Q = {}
    states = grid.all_states()
    for s in states:
        Q[s] = {}
        for a in ALL_POSSIBLE_ACTIONS:
            Q[s][a] = 0

    # let's also keep track of how many times Q[s] has been updated
    update_counts = {}
    update_counts_sa = {}
    for s in states:
        update_counts_sa[s] = {}
        for a in ALL_POSSIBLE_ACTIONS:
            update_counts_sa[s][a] = 1.0

    # repeat until convergence
    t = 1.0
    deltas = []
    for it in range(5000):
        if it % 100 == 0:
            t += 1e-2
        #if it % 2000 == 0:
            #print "it:", it

        # instead of 'generating' an epsiode, we will PLAY
        # an episode within this loop
        s = (3, 0) # start state
        grid.set_state(s)

        # the first (s, r) tuple is the state we start in and 0
        # (since we don't get a reward) for simply starting the game
        # the last (s, r) tuple is the terminal state and the final reward
        # the value for the terminal state is by definition 0, so we don't
        # care about updating it.
        #a = max_dict(Q[s])[0]
        a = np.random.choice(ALL_POSSIBLE_ACTIONS)
        biggest_change = 0
        #print 'old a is', a
        while not grid.game_over():
            r = grid.move(a)
            s2 = grid.current_state()
            #print ' a  befor update is ::', a, s2

            # we need the next action as well since Q(s,a) depends on Q(s',a')
            # if s2 not in policy then it's a terminal state, all Q are 0

            a2 = max_dict(Q[s2])[0]
            #print 'a2 is max dic Qs before random::', a2
            a2 = random_action(a2, eps=1) # epsilon-greedy
            #print 'a2 is', a2

            #a2 = np.random.choice(ALL_POSSIBLE_ACTIONS)
            # we will update Q(s,a) AS we experience the episode
            #alpha = ALPHA / update_counts_sa[s][a]
            alpha = 0.1
            update_counts_sa[s][a] += 0.005
            old_qsa = Q[s][a]
            Q[s][a] = Q[s][a] + alpha*(r + GAMMA*Q[s2][a2] - Q[s][a])

            biggest_change = max(biggest_change, np.abs(old_qsa - Q[s][a]))

            # we would like to know how often Q(s) has been updated too
            update_counts[s] = update_counts.get(s,0) + 1

            # next state becomes current state
            s = s2
            a = a2

            #print ' every time updated a is ', a, s,Q[s], Q[s][a]

        deltas.append(biggest_change)

    plt.plot(deltas)
    plt.show()

    # determine the policy from Q*
    # find V* from Q*
    policy = {}
    V = {}
    for s in grid.actions.keys():
        a, max_q = max_dict(Q[s])
        policy[s] = a
        V[s] = max_q
    #print ("policy and v", policy[s], V[s])
    # what's the proportion of time we spend updating each part of Q?
    #print "update counts:"
    total =sum(update_counts.values())
    for k, v in update_counts.items():
        update_counts[k] = float(v) / total
    #print_values(update_counts, grid)

    print ("values:")
    print_values(V, grid)
    print ("policy:")
    print_policy(policy, grid)

    """
    for s in policy.keys():
        print 's is', s, 'Q is ', Q[s]['R']
        print 's is', s, 'Q is ', Q[s]['L']
        print 's is', s, 'Q is ', Q[s]['U']
        print 's is', s, 'Q is ', Q[s]['D']
    """
